namespace Syntriax.Engine.Core;

/// <summary>
/// Represents an HSV color.
/// </summary>
/// <param name="hue">Hue of the <see cref="ColorHSV"/>.</param>
/// <param name="saturation">Saturation of the <see cref="ColorHSV"/>.</param>
/// <param name="value">Value of the <see cref="ColorHSV"/>.</param>
/// <remarks>
/// Initializes a new instance of the <see cref="ColorHSV"/> struct with the specified values.
/// </remarks>
[System.Diagnostics.DebuggerDisplay("{ToString(),nq}")]
public readonly struct ColorHSV(float hue, float saturation, float value)
{
    /// <summary>
    /// The Hue value of the <see cref="ColorHSV"/>.
    /// </summary>
    public readonly float Hue = hue.Clamp(0f, 1f);

    /// <summary>
    /// The Saturation value of the <see cref="ColorHSV"/>.
    /// </summary>
    public readonly float Saturation = saturation.Clamp(0f, 1f);

    /// <summary>
    /// The Value value of the <see cref="ColorHSV"/>.
    /// </summary>
    public readonly float Value = value.Clamp(0f, 1f);

    public static ColorHSV operator -(ColorHSV color) => new(color.Hue.OneMinus().Clamp(0f, 1f), color.Saturation.OneMinus().Clamp(0f, 1f), color.Value.OneMinus().Clamp(0f, 1f));
    public static ColorHSV operator +(ColorHSV left, ColorHSV right) => new((left.Hue + right.Hue).Clamp(0f, 1f), (left.Saturation + right.Saturation).Clamp(0f, 1f), (left.Value + right.Value).Clamp(0f, 1f));
    public static ColorHSV operator -(ColorHSV left, ColorHSV right) => new((left.Hue - right.Hue).Clamp(0f, 1f), (left.Saturation - right.Saturation).Clamp(0f, 1f), (left.Value - right.Value).Clamp(0f, 1f));
    public static ColorHSV operator *(ColorHSV left, ColorHSV right) => new((left.Hue * right.Hue).Clamp(0f, 1f), (left.Saturation * right.Saturation).Clamp(0f, 1f), (left.Value * right.Value).Clamp(0f, 1f));
    public static ColorHSV operator *(ColorHSV color, float value) => new((color.Hue * value).Clamp(0f, 1f), (color.Saturation * value).Clamp(0f, 1f), (color.Value * value).Clamp(0f, 1f));
    public static ColorHSV operator *(float value, ColorHSV color) => new((color.Hue * value).Clamp(0f, 1f), (color.Saturation * value).Clamp(0f, 1f), (color.Value * value).Clamp(0f, 1f));
    public static ColorHSV operator /(ColorHSV color, float value) => new((color.Hue / value).Clamp(0f, 1f), (color.Saturation / value).Clamp(0f, 1f), (color.Value / value).Clamp(0f, 1f));
    public static bool operator ==(ColorHSV left, ColorHSV right) => left.Hue == right.Hue && left.Saturation == right.Saturation && left.Value == right.Value;
    public static bool operator !=(ColorHSV left, ColorHSV right) => left.Hue != right.Hue || left.Saturation != right.Saturation || left.Value != right.Value;

    public static implicit operator ColorHSV(ColorRGBA rgba) => (ColorRGB)rgba;
    public static implicit operator ColorHSV(ColorRGB rgb)
    {
        float hue;
        float saturation;
        float value;

        float rd = rgb.R / 255f;
        float gd = rgb.G / 255f;
        float bd = rgb.B / 255f;

        float max = Math.Max(rd, Math.Max(gd, bd));
        float min = Math.Min(rd, Math.Min(gd, bd));
        float delta = max - min;

        if (delta.ApproximatelyEquals(0))
            hue = 0f;
        else if (max.ApproximatelyEquals(rd))
            hue = 60f * ((gd - bd) / delta % 6f);
        else if (max.ApproximatelyEquals(gd))
            hue = 60f * (((bd - rd) / delta) + 2f);
        else
            hue = 60f * (((rd - gd) / delta) + 4f);

        if (hue < 0f)
            hue += 360f;

        hue /= 360f;
        saturation = max.ApproximatelyEquals(0f) ? 0f : delta / max;
        value = max;

        return new(hue, saturation, value);
    }

    /// <summary>
    /// Inverts the given <see cref="ColorHSV"/>.
    /// </summary>
    /// <param name="color">The <see cref="ColorHSV"/>.</param>
    /// <returns>The inverted <see cref="ColorHSV"/>.</returns>
    public static ColorHSV Invert(ColorHSV color) => -color;

    /// <summary>
    /// Adds two <see cref="ColorHSV"/>s.
    /// </summary>
    /// <param name="left">The first <see cref="ColorHSV"/>.</param>
    /// <param name="right">The second <see cref="ColorHSV"/>.</param>
    /// <returns>The sum of the two <see cref="ColorHSV"/>s.</returns>
    public static ColorHSV Add(ColorHSV left, ColorHSV right) => left + right;

    /// <summary>
    /// Subtracts one <see cref="ColorHSV"/> from another.
    /// </summary>
    /// <param name="left">The <see cref="ColorHSV"/> to subtract from.</param>
    /// <param name="right">The <see cref="ColorHSV"/> to subtract.</param>
    /// <returns>The result of subtracting the second <see cref="ColorHSV"/> from the first.</returns>
    public static ColorHSV Subtract(ColorHSV left, ColorHSV right) => left - right;

    /// <summary>
    /// Multiplies a <see cref="ColorHSV"/> by a scalar value.
    /// </summary>
    /// <param name="color">The <see cref="ColorHSV"/>.</param>
    /// <param name="value">The scalar value.</param>
    /// <returns>The result of multiplying the <see cref="ColorHSV"/> by the scalar value.</returns>
    public static ColorHSV Multiply(ColorHSV color, float value) => color * value;

    /// <summary>
    /// Divides a <see cref="ColorHSV"/> by a scalar value.
    /// </summary>
    /// <param name="color">The <see cref="ColorHSV"/>.</param>
    /// <param name="value">The scalar value.</param>
    /// <returns>The result of dividing the <see cref="ColorHSV"/> by the scalar value.</returns>
    public static ColorHSV Divide(ColorHSV color, float value) => color / value;

    /// <summary>
    /// Calculates the <see cref="ColorHSV"/> from one point to another.
    /// </summary>
    /// <param name="from">The starting point.</param>
    /// <param name="to">The ending point.</param>
    /// <returns>The <see cref="ColorHSV"/> from the starting point to the ending point.</returns>
    public static ColorHSV FromTo(ColorHSV from, ColorHSV to) => to - from;

    /// <summary>
    /// Performs linear interpolation between two <see cref="ColorHSV"/>s.
    /// </summary>
    /// <param name="from">The starting <see cref="ColorHSV"/> (t = 0).</param>
    /// <param name="to">The ending <see cref="ColorHSV"/> (t = 1).</param>
    /// <param name="t">The interpolation parameter.</param>
    /// <returns>The interpolated <see cref="ColorHSV"/>.</returns>
    public static ColorHSV Lerp(ColorHSV from, ColorHSV to, float t) => from + FromTo(from, to) * t;

    /// <summary>
    /// Checks if two <see cref="ColorHSV"/>s are approximately equal within a specified epsilon range.
    /// </summary>
    /// <param name="left">The first <see cref="ColorHSV"/>.</param>
    /// <param name="right">The second <see cref="ColorHSV"/>.</param>
    /// <param name="epsilon">The epsilon range.</param>
    /// <returns><see cref="true"/> if the <see cref="ColorHSV"/>s are approximately equal; otherwise, <see cref="false"/>.</returns>
    public static bool ApproximatelyEquals(ColorHSV left, ColorHSV right, float epsilon = float.Epsilon)
        => left.Hue.ApproximatelyEquals(right.Hue, epsilon) && left.Saturation.ApproximatelyEquals(right.Saturation, epsilon) && left.Value.ApproximatelyEquals(right.Value, epsilon);

    /// <summary>
    /// Determines whether the specified object is equal to the current <see cref="ColorHSV"/>.
    /// </summary>
    /// <param name="obj">The object to compare with the current <see cref="ColorHSV"/>.</param>
    /// <returns><see cref="true"/> if the specified object is equal to the current <see cref="ColorHSV"/>; otherwise, <see cref="false"/>.</returns>
    public override bool Equals(object? obj) => obj is ColorHSV colorHSV && this == colorHSV;

    /// <summary>
    /// Generates a hash code for the <see cref="ColorHSV"/>.
    /// </summary>
    /// <returns>A hash code for the <see cref="ColorHSV"/>.</returns>
    public override int GetHashCode() => System.HashCode.Combine(Hue, Saturation, Value);

    /// <summary>
    /// Converts the <see cref="ColorHSV"/> to its string representation.
    /// </summary>
    /// <returns>A string representation of the <see cref="ColorHSV"/>.</returns>
    public override string ToString() => $"{nameof(ColorHSV)}({Hue}, {Saturation}, {Value})";
}

/// <summary>
/// Provides extension methods for <see cref="ColorHSV"/> type.
/// </summary>
public static class ColorHSVExtensions
{
    /// <inheritdoc cref="ColorHSV.Add(ColorHSV, ColorHSV)" />
    public static ColorHSV Add(this ColorHSV color, ColorHSV value) => ColorHSV.Add(color, value);

    /// <inheritdoc cref="ColorHSV.Subtract(ColorHSV, ColorHSV)" />
    public static ColorHSV Subtract(this ColorHSV color, ColorHSV value) => ColorHSV.Subtract(color, value);

    /// <inheritdoc cref="ColorHSV.Multiply(ColorHSV, ColorHSV)" />
    public static ColorHSV Multiply(this ColorHSV color, float value) => ColorHSV.Multiply(color, value);

    /// <inheritdoc cref="ColorHSV.Divide(ColorHSV, ColorHSV)" />
    public static ColorHSV Divide(this ColorHSV color, float value) => ColorHSV.Divide(color, value);

    /// <inheritdoc cref="ColorHSV.FromTo(ColorHSV, ColorHSV)" />
    public static ColorHSV FromTo(this ColorHSV from, ColorHSV to) => ColorHSV.FromTo(from, to);

    /// <inheritdoc cref="ColorHSV.Lerp(ColorHSV, ColorHSV, float)" />
    public static ColorHSV Lerp(this ColorHSV from, ColorHSV to, float t) => ColorHSV.Lerp(from, to, t);

    /// <inheritdoc cref="ColorHSV.ApproximatelyEquals(ColorHSV, ColorHSV, float) " />
    public static bool ApproximatelyEquals(this ColorHSV left, ColorHSV right, float epsilon = float.Epsilon) => ColorHSV.ApproximatelyEquals(left, right, epsilon);
}
